import mimetypes
import os

from django.utils.functional import cached_property


class PrivateFile(object):
    """
    A wrapper object that describes the file that is being accessed.
    """

    def __init__(self, request, storage, relative_name):
        self.request = request
        self.storage = storage
        self.relative_name = relative_name

    @cached_property
    def full_path(self):
        # Not using self.storage.open() as the X-Sendfile needs a normal path.
        return self.storage.path(self.relative_name)

    def exists(self):
        return os.path.exists(self.full_path)

    @cached_property
    def content_type(self):
        """
        Return the HTTP ``Content-Type`` header value for a filename.
        """
        mimetype, encoding = mimetypes.guess_type(self.full_path)
        return mimetype or 'application/octet-stream'



"""
Sending files efficiently for different kind of webservers.
"""
import os

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse
from django.utils.lru_cache import lru_cache
from django.views.static import serve

from .utils import import_symbol


@lru_cache()
def get_server_class(path):
    if '.' in path:
        return import_symbol(path)
    elif path == 'django':
        return DjangoServer()
    elif path == 'apache':
        return ApacheXSendfileServer
    elif path == 'nginx':
        return NginxXAccelRedirectServer
    else:
        raise ImproperlyConfigured(
            "PRIVATE_STORAGE_SERVER setting should be 'nginx', 'apache', 'django' or a python class path."
        )


class DjangoServer(object):
    """
    Serve static files from the local filesystem through Django.
    This is a bad idea for most situations other than testing.
    The file data is copied multiple times; read by Django, written to WSGI,
    read by webserver, outputted to client.
    """

    @staticmethod
    def serve(private_file):
        # This supports If-Modified-Since and sends the file in 8KB chunks
        return serve(private_file.request, private_file.full_path, document_root='/', show_indexes=False)


class ApacheXSendfileServer(object):
    """
    Serve files for Apache with ``X-Sendfile``.
    """

    @staticmethod
    def serve(private_file):
        response = HttpResponse()
        response['X-Sendfile'] = private_file.full_path
        response['Content-Type'] = private_file.content_type
        return response


class NginxXAccelRedirectServer(object):
    """
    Serve the files for Nginx with ``X-Accel-Redirect``.
    Add the following configuration::

        location /private-x-accel-redirect/ (
            internal;
            alias /home/user/my/path/to/private/media/;
        )

    Or update the ``PRIVATE_STORAGE_INTERNAL_URL`` setting to use a different URL prefix.
    """

    @staticmethod
    def serve(private_file):
        internal_url = os.path.join(settings.PRIVATE_STORAGE_INTERNAL_URL, private_file.relative_name)
        response = HttpResponse()
        response['X-Accel-Redirect'] = internal_url
        response['Content-Type'] = private_file.content_type
        return response



"""
Views to send private files.
"""
from django.http import HttpResponseForbidden, Http404
from django.views.generic import View

from . import appconfig
from .models import PrivateFile
from .servers import get_server_class
from .storage import private_storage
from .utils import import_symbol


class PrivateStorageView(View):
    """
    Return the uploaded files
    """

    #: The storage class to retrieve files from
    storage = private_storage

    #: The authorisation rule for accessing
    can_access_file = import_symbol(appconfig.PRIVATE_STORAGE_AUTH_FUNCTION)

    #: Import the server class once
    server_class = get_server_class(appconfig.PRIVATE_STORAGE_SERVER)

    def get(self, request, *args, **kwargs):
        """
        Handle incoming GET requests
        """
        # Wrap all relevant data in a single object,
        # so this is easy to extend and server implementations can pick what they need.
        private_file = PrivateFile(
            request=self.request,
            storage=self.storage,
            relative_name=self.kwargs['path']
        )

        if not self.can_access_file(private_file):
            return HttpResponseForbidden('Private storage access denied')

        if not private_file.exists():
            raise Http404("File not found")

        return self.serve_file(private_file)

    def serve_file(self, private_file):
        """
        Serve the file that was retrieved from the storage.
        The relative path can be found in ``self.kwargs['path']``.
        """
        return self.server_class().serve(private_file)



"""
Django Storage interface
"""
from django.core.files.storage import FileSystemStorage
from django.core.urlresolvers import reverse_lazy
from . import appconfig

__all__ = (
    'private_storage',
    'PrivateStorage',
)


class PrivateStorage(FileSystemStorage):
    """
    Interface to the Django storage system,
    storing the files in a private folder.
    """
    def __init__(self, location=None, base_url=None, **kwargs):
        if location is None:
            location = appconfig.PRIVATE_STORAGE_ROOT

        super(PrivateStorage, self).__init__(
            location=location,
            base_url=base_url,
            **kwargs
        )
        if base_url is None:
            # When base_url is not given, it's autodetected.
            # However, as the super method checks for base_url.endswith('/'),
            # the attribute is overwritten here to avoid breaking lazy evaluation.
            self.base_url = reverse_lazy('serve_private_file', kwargs={'path': ''})


# Singleton instance.
private_storage = PrivateStorage()



from django.core.exceptions import ImproperlyConfigured

try:
    from importlib import import_module  # Python 2.7+
except ImportError:
    from django.utils.importlib import import_module


def import_symbol(import_path):
    """
    Import a class or attribute by name.
    """
    try:
        dot = import_path.rindex('.')
    except ValueError:
        raise ImproperlyConfigured("{0} isn't a Python path.".format(import_path))

    module, classname = import_path[:dot], import_path[dot + 1:]
    try:
        mod = import_module(module)
    except ImportError as e:
        raise ImproperlyConfigured('Error importing module {0}: "{1}"'.format(module, e))

    try:
        return getattr(mod, classname)
    except AttributeError:
        raise ImproperlyConfigured('Module "{0}" does not define a "{1}" class.'.format(module, classname))
